/*
 * Copyright (C) 2008 Petr Hruska <phru@ucw.cz> 
 * 
 * This file is part of Cutter module.
 * 
 * Cutter module is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <stdlib.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>
#include <stdio.h>

#include "../../mem/mem.h"
#include "billing.h"
#include "bparser.h"
#include "misc.h"

#define COMMAND_BUFFER_SIZE 512

/*
 * Instances of module parameters
 * */

/* billing socket string */
char *billing_address = 0; 
int billing_timeout = 10; 
int billing_duration = -1; 
char* billing_gateway = NULL; 

/*
 * Global variables
 * */
const char const *default_billing_host = "127.0.0.1";
const int default_billing_port = 9024;

struct sockaddr_in billing_addr;
int billing_socket = -1;

int setup_billing_addr(char const* host, int n_port)
{
	struct hostent* he;

	he = gethostbyname(host);
	if (!he || !he->h_addr_list || !he->h_addr_list[0]) {
		LOG(L_ERR, "ERROR: cutter: setup_billing_addr: "
		"can't resolve host name \"%s\"\n", host);
		return -1;
	}

	if (he->h_addrtype != AF_INET) {
		LOG(L_ERR, "ERROR: cutter: setup_billing_addr: "
		"unsupported address family %d\n", he->h_addrtype);
		return -1;
	}

	memcpy(&billing_addr.sin_addr, he->h_addr_list[0], he->h_length);
	billing_addr.sin_family=AF_INET;
	billing_addr.sin_port=htons(n_port);

	return 0;
}


/* Parse given parameter to host name and port number. Use
 * defult_billing_address when null is passed and default_billing_port when
 * port number is missing. Try to resolve host part and setup billing_addr.
 */
int parse_billing_address(char const* billing_address)
{
	char* endptr;
	char const* port;
	char* host;
	int n_port, host_len;
	int r;

	if (!billing_address) {
		LOG(L_NOTICE, "NOTICE: cutter: parse_billing_address: "
		"Using default billing address %s:%d\n", 
		default_billing_host, default_billing_port);

		return setup_billing_addr(default_billing_host, 
			default_billing_port);
	} else {

		port = strchr(billing_address, ':');
		if (port) {

			host_len = port - billing_address;
			port++;
			n_port = strtol(port, &endptr, 0);

			if (!*port || *endptr || n_port < 0 
			|| n_port > 65535) {

				LOG(L_ERR, 
				"ERROR: cutter: parse_billing_address: "
				"invalid port specification\n");

				return -1;
			}

		} else {
			host_len = strlen(billing_address);

			LOG(L_NOTICE, "NOTICE: cutter: parse_billing_address: "
			"Using default billing port %d\n", 
			default_billing_port);

			n_port = default_billing_port;
		}
		
		host = pkg_malloc(host_len + 1);
		if (!host) {
			LOG(L_ERR, "ERROR: cutter: parse_billing_address: "
			"memory allocation failed.\n");
			return -1;
		}
		memcpy(host, billing_address, host_len);
		host[host_len] = 0;
		r = setup_billing_addr(host, n_port);
		pkg_free(host);

		return r;
	}
}


void billing_close_socket()
{
	close(billing_socket);
	billing_socket = -1;
}



/*
 * Create socket, connect to an address specified in billing_addr.
 */
int billing_connect()
{
	int r;

	DBG("DEBUG: cutter: billing_connect: start\n");
	billing_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (billing_socket == -1) {
		LOG(L_ERR, "ERROR: cutter: billing_connect: "
		"Can't create socket.\n");
		return -1;
	}

	r = connect(billing_socket, (struct sockaddr*) &billing_addr, 
		sizeof(billing_addr));
	if (r) {
		LOG(L_ERR, "ERROR: cutter: billing_connect: "
		"Can't connect socket.\n");
		return -1;
	}

	r = fcntl(billing_socket, F_SETFL, O_NONBLOCK);
	if (r) {
		LOG(L_ERR, "ERROR: cutter: billing_connect: "
		"Can't set socket nonblocking.\n");
		return -1;
	}

	return 0;
}


int billing_write_command(char const* cmd)
{
	int len, n, r;
	struct pollfd pfd;
	int timeout = billing_timeout*1000;

	pfd.fd = billing_socket;
	pfd.events = POLLOUT;
	DBG("DEBUG: cutter: billing_connect: writing '%s' to fd %d\n.", 
	cmd, billing_socket);

	len = strlen(cmd);
	while (len >= 0) {
		r = poll(&pfd, 1, timeout);
		if (r == -1) {
			LOG(L_ERR, "ERROR: cutter: billing_write_command: "
			"poll failed.\n");
			return -1;
		}

		if (r == 0) {
			LOG(L_ERR, "ERROR: cutter: billing_write_command: "
			"timeout while writing to billing server.\n");
			return -1;
		}

		if (len > 0)
			n = write(billing_socket, cmd, len);
		else
			n = write(billing_socket, "\n", 1);

		// should not write zero bytes after successfull poll
		if (n <= 0) {
			LOG(L_WARN, "WARNING: cutter: billing_write_command: "
			"Write failure, billing server restarted?.\n");
			return -1;
		}

		len -= n;
		cmd += n;
	}

	return 0;
}


/* 
 * Send a command to billing socket. Adds terminating newline character. If
 * there is an error onsocket, try reconnect and send again. If it does not
 * help, return an error code.
 *
 * This behaviour should be usefull when billing server is restarted.
 */
int billing_command(char* cmd)
{
	if (billing_socket == -1) {
		DBG("DEBUG: cutter: billing_command: "
		"connection not established yet, connecting ...\n");
		if (billing_connect() < 0) {
			LOG(L_ERR, "ERROR: cutter: billing_command: "
			"connection failed.\n");
			return -1;
		}
	}
	if (billing_write_command(cmd) < 0) {
		LOG(L_WARN, "WARNING: cutter: billing_command: "
		"Write failure, trying reconnect and write again.\n");
		billing_close_socket();
		if (billing_connect() < 0) {
			LOG(L_ERR, "ERROR: cutter: billing_command: "
			"Reconnection failed, giving up.\n");
			return -1;
		}
		if (billing_write_command(cmd) < 0) {
			LOG(L_ERR, "ERROR: cutter: billing_command: "
			"Can't write even after reconnection - strange.\n");
			return -1;
		}
	}
	
	return 0;
}

/*
 * Read an answer which ends with '\n'. Return -2 on overflow, -1 on
 * other errors, number of read bytes (including terminating characters) on
 * success.
 */
int billing_readline(char* buffer, int buflen)
{
	int n, pos, r;
	struct pollfd pfd;
	int timeout = billing_timeout*1000;

	pfd.fd = billing_socket;
	pfd.events = POLLIN;

	pos = 0; //buffer[pos] is the first dirty (not read yet) byte

	while (pos < buflen) {
		r = poll(&pfd, 1, timeout);
		if (r == -1) {
			LOG(L_ERR, "ERROR: cutter: billing_readline(): "
			"poll failed.\n");
			return -1;
		}

		if (r == 0) {
			LOG(L_ERR, "ERROR: cutter: billing_readline(): "
			"timeout while reading from billing server.\n");
			return -1;
		}

		n = read(billing_socket, buffer + pos, buflen - pos);
		// Should not read zero bytes after successfull poll.
		// Maybe signal could iterrupt read(), but I believe
		// it's very rare condition.
		if (n <= 0) {
			LOG(L_WARN, "WARNING: cutter: billing_readline(): "
			"Read failure, interrupt during read()?.\n");
			return -1;
		}

		pos += n;
		// Check the terminating character. There should be only one
		// message ending with \n from server.

		if (pos >= 1 && buffer[pos - 1] == '\n')
			return pos;
	}

	LOG(L_ERR, "ERROR: cutter: billing_readline(): Buffer overflow.\n");
	return -2;
}

/* dest = dest + s */
static inline int addstr2(str *dest, int dest_max_len, char const* s, int s_len)
{
	if (dest->len + s_len > dest_max_len) return -1;

	memcpy(dest->s + dest->len, s, s_len);
	dest->len += s_len;
	return 0;
}

static inline int addstr(str *dest, int dest_max_len, str const* s)
{
	return addstr2(dest, dest_max_len, s->s, s->len);
}

static inline int addstr1(str *dest, int dest_max_len, char const* s)
{
	int len = strlen(s);
	return addstr2(dest, dest_max_len, s, len);
}


static inline int buffer_overflow()
{
	LOG(L_ERR, "ERROR: cutter: buffer_overflow(): buffer overflow.\n");
	return -1;
}



void* billing_alloc()
{
	str *p;
	p = pkg_malloc(sizeof(*p));

	if (!p) return NULL;

	p->s = pkg_malloc(COMMAND_BUFFER_SIZE);
	if (!p->s) {
		LOG(L_ERR, "ERROR: cutter: billing_alloc(): out of memory.\n");
		pkg_free(p);
		return NULL;
	}

	p->len = COMMAND_BUFFER_SIZE;
	return p;
}



void billing_free(void* buf)
{
	str* p = (str*) buf;
	if (!p) {
		LOG(L_WARN, "WARNING: cutter: billing_free(): NULL value of command buffer.\n");
		return;
	}

	if (!p->s) 
		LOG(L_WARN, "WARNING: cutter: billing_free(): NULL value of previously allocated string.\n");
	else
		pkg_free(p->s);

	pkg_free(p);
}



/*
 * Send command like
 *
 * Calling Callid="372c711d0e90c2015ddfc54746234646@127.0.0.1-as377214c4",
 * Caller="951018001", Callee="951018001"\n
 * 
 * to billing server and read answer.
 *
 * Return 0 on success, -1 on failure.
 */
int billing_calling(void* bbuf, str const* callid, str const* caller, str const* callee,
const str* srcip, str* callcon, time_t* time, int* monitor, int* result)
{
	const int buflen = ((str*) bbuf)->len;

	str buf;
	buf.s = ((str*) bbuf)->s;
	buf.len = 0;

	/* Callid */
	if (addstr1(&buf, buflen, "Calling Callid=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callid) == -1)
		return buffer_overflow();

	/* Caller */
	if (addstr1(&buf, buflen, "\", Caller=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, caller) == -1)
		return buffer_overflow();

	/* Callee */
	if (addstr1(&buf, buflen, "\", Callee=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callee) == -1)
		return buffer_overflow();
	
	/* SrcIP */
	if (addstr1(&buf, buflen, "\", SrcIP=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, srcip) == -1)
		return buffer_overflow();

	/* terminating '"' and null */
	if (addstr2(&buf, buflen, "\"", 2) == -1) return buffer_overflow();


	if (billing_command(buf.s) < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_calling(): "
		"write command failed.\n");
		return -1;
	}


	/*
	 * message send, wait for reply
	 * */


	int n = billing_readline(buf.s, buflen);
	if (n < 1) {
		LOG(L_ERR, "ERROR: cutter: billing_calling(): "
		"result reading failed.\n");
		return -1;
	}

	// should read at least two terminating characters
	buf.s[--n] = 0;
	DBG("DEBUG: cutter: billing_calling(): received '%s'.\n", buf.s);


	int code = bparser_code(buf.s);
	if (code < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_calling(): "
		"Invalid format of answer code.\n");
		return -1;
	}
	DBG("DEBUG: cutter: billing_calling(): code = '%d'.\n", code);
	*result = code;

	char* msg = buf.s + 3; // 200 Descr=...

	// default values
	*monitor = 0;
	*time = 0;
	callcon->len = 0;
	callcon->s = NULL;


	str name, val;
	int pos = 0;
	int index;
	static char const* params[5] = {"Descr", "Callcon", "Time", "Monitor", NULL};

	int r = bparser_parse(msg, params, &name, &val, &pos, &index);
	while (r == 1) {
		//DBG("DEBUG: cutter: billing_calling(): bparser index = %d.\n", index);
		
		switch (index) {
		case -1:
			LOG(L_WARN, "WARNING: cutter: billing_calling(): "
			"unknown parameter in reply.\n");
			break;
		case 1: //Callcon
			*callcon = val;
			break;
		case 2: //Time
			*time = str2time_t(val);
			break;
		case 3: //Monitor
			*monitor = str2signed_int(val);
			break;
		}
		r = bparser_parse(msg, params, &name, &val, &pos, &index);
	}
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_calling(): "
		"invalid syntax of reply.\n");
		return -1;
	}

	if (billing_command("200 Descr=\"OK\"\n") < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_calling(): "
		"write ack command failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: billing_calling(): billing code = %d, callcon = %.*s time = %ld, monitor=%d.\n", 
		*result, callcon->len, callcon->s, *time, *monitor);

	return 0;
}


int billing_answered(void* bbuf, str const* callid, time_t time, time_t *next_continue,
int* monitor, int* result)
{
	const int buflen = ((str*) bbuf)->len;

	str buf;
	buf.s = ((str*) bbuf)->s;
	buf.len = 0;

	/* Callid */
	if (addstr1(&buf, buflen, "Answered Callid=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callid) == -1)
		return buffer_overflow();

	/* Time */
	int r = snprintf(buf.s + buf.len, buflen - buf.len,
	"\", Time=%ld", time);
	if (r >= buflen - buf.len) return buffer_overflow();
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"snprintf output error.\n");
		return -1;
	}
	buf.len += r;

	// snprintf() adds terminating null

	if (billing_command(buf.s) < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"write command failed.\n");
		return -1;
	}


	/*
	 * message send, wait for reply
	 * */


	int n = billing_readline(buf.s, buflen);
	if (n < 1) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"result reading failed.\n");
		return -1;
	}

	// should read at least two terminating characters
	buf.s[--n] = 0;
	DBG("DEBUG: cutter: billing_answered(): received '%s'.\n", buf.s);


	int code = bparser_code(buf.s);
	if (code < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"Invalid format of answer code.\n");
		return -1;
	}
	*result = code;

	char* msg = buf.s + 3; // 200 Descr=...

	// default values
	*monitor = 0;
	*next_continue = 0;


	str name, val;
	int pos = 0;
	int index;
	static char const* params[4] = {"Descr", "Time", "Monitor", NULL};

	r = bparser_parse(msg, params, &name, &val, &pos, &index);
	while (r == 1) {
		
		switch (index) {
		case -1:
			LOG(L_WARN, "WARNING: cutter: billing_answered(): "
			"unknown parameter in reply.\n");
			break;
		case 1: //Time
			*next_continue = str2time_t(val);
			break;
		case 2: //Monitor
			*monitor = str2signed_int(val);
			break;
		}
		r = bparser_parse(msg, params, &name, &val, &pos, &index);
	}
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"invalid syntax of reply.\n");
		return -1;
	}

	if (billing_command("200 Descr=\"OK\"\n") < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_answered(): "
		"write ack command failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: billing_answered(): billing result = %d, time = %ld, monitor=%d.\n", *result, *next_continue, *monitor);

	return 0;
}


int billing_cancelled(void* bbuf, str const* callid, time_t time, str const* description,
int* result)
{
	const int buflen = ((str*) bbuf)->len;

	str buf;
	buf.s = ((str*) bbuf)->s;
	buf.len = 0;

	/* Callid */
	if (addstr1(&buf, buflen, "Cancelled Callid=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callid) == -1)
		return buffer_overflow();

	/* Time */
	int r = snprintf(buf.s + buf.len, buflen - buf.len,
	"\", Time=%ld", time);
	if (r >= buflen - buf.len) return buffer_overflow();
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"snprintf output error.\n");
		return -1;
	}
	buf.len += r;

	/* Description  */
	if (addstr1(&buf, buflen, ", Descr=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, description) == -1)
		return buffer_overflow();

	/* terminating '"' and null */
	if (addstr2(&buf, buflen, "\"", 2) == -1) return buffer_overflow();


	if (billing_command(buf.s) < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"write command failed.\n");
		return -1;
	}


	/*
	 * message send, wait for reply
	 * */


	int n = billing_readline(buf.s, buflen);
	if (n < 1) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"result reading failed.\n");
		return -1;
	}

	// should read at least two terminating characters
	buf.s[--n] = 0;
	DBG("DEBUG: cutter: billing_cancelled(): received '%s'.\n", buf.s);


	int code = bparser_code(buf.s);
	if (code < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"Invalid format of answer code.\n");
		return -1;
	}
	*result = code;

	char* msg = buf.s + 3; // 200 Descr=...


	str name, val;
	int pos = 0;
	int index;
	static char const* params[2] = {"Descr", NULL};

	r = bparser_parse(msg, params, &name, &val, &pos, &index);
	while (r == 1) {
		
		switch (index) {
		case -1:
			LOG(L_WARN, "WARNING: cutter: billing_cancelled(): "
			"unknown parameter in reply.\n");
			break;
		}
		r = bparser_parse(msg, params, &name, &val, &pos, &index);
	}
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"invalid syntax of reply.\n");
		return -1;
	}

	if (billing_command("200 Descr=\"OK\"\n") < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_cancelled(): "
		"write ack command failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: billing_cancelled(): billing result = %d.\n", *result);

	return 0;
}


int billing_continue(void* bbuf, str const* callid, time_t *next_continue,
int* monitor, int* result)
{
	const int buflen = ((str*) bbuf)->len;

	str buf;
	buf.s = ((str*) bbuf)->s;
	buf.len = 0;

	/* Callid */
	if (addstr1(&buf, buflen, "Continue Callid=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callid) == -1)
		return buffer_overflow();

	/* terminating '"' and null */
	if (addstr2(&buf, buflen, "\"", 2) == -1) return buffer_overflow();

	if (billing_command(buf.s) < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_continue(): "
		"write command failed.\n");
		return -1;
	}


	/*
	 * message send, wait for reply
	 * */


	int n = billing_readline(buf.s, buflen);
	if (n < 1) {
		LOG(L_ERR, "ERROR: cutter: billing_continue(): "
		"result reading failed.\n");
		return -1;
	}

	// should read at least two terminating characters
	buf.s[--n] = 0;
	DBG("DEBUG: cutter: billing_continue(): received '%s'.\n", buf.s);


	int code = bparser_code(buf.s);
	if (code < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_continue(): "
		"Invalid format of answer code.\n");
		return -1;
	}
	*result = code;

	char* msg = buf.s + 3; // 200 Descr=...

	// default values
	*monitor = 0;
	*next_continue = 0;


	str name, val;
	int pos = 0;
	int index;
	static char const* params[4] = {"Descr", "Time", "Monitor", NULL};

	int r = bparser_parse(msg, params, &name, &val, &pos, &index);
	while (r == 1) {
		
		switch (index) {
		case -1:
			LOG(L_WARN, "WARNING: cutter: billing_continue(): "
			"unknown parameter in reply.\n");
			break;
		case 1: //Time
			*next_continue = str2time_t(val);
			break;
		case 2: //Monitor
			*monitor = str2signed_int(val);
			break;
		}
		r = bparser_parse(msg, params, &name, &val, &pos, &index);
	}
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_continue(): "
		"invalid syntax of reply.\n");
		return -1;
	}

	if (billing_command("200 Descr=\"OK\"\n") < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_continue(): "
		"write ack command failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: billing_continue(): billing result = %d, time = %ld, monitor=%d.\n", *result, *next_continue, *monitor);

	return 0;
}


int billing_hangedup(void* bbuf, str const* callid, time_t time, str const* description,
int* result)
{
	const int buflen = ((str*) bbuf)->len;

	str buf;
	buf.s = ((str*) bbuf)->s;
	buf.len = 0;

	/* Callid */
	if (addstr1(&buf, buflen, "HangedUp Callid=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, callid) == -1)
		return buffer_overflow();

	/* Time */
	int r = snprintf(buf.s + buf.len, buflen - buf.len,
	"\", Time=%ld", time);
	if (r >= buflen - buf.len) return buffer_overflow();
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"snprintf output error.\n");
		return -1;
	}
	buf.len += r;

	/* Description  */
	if (addstr1(&buf, buflen, ", Descr=\"") == -1)
		return buffer_overflow();
	if (addstr(&buf, buflen, description) == -1)
		return buffer_overflow();

	/* terminating '"' and null */
	if (addstr2(&buf, buflen, "\"", 2) == -1) return buffer_overflow();


	if (billing_command(buf.s) < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"write command failed.\n");
		return -1;
	}


	/*
	 * message send, wait for reply
	 * */


	int n = billing_readline(buf.s, buflen);
	if (n < 1) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"result reading failed.\n");
		return -1;
	}

	// should read at least two terminating characters
	buf.s[--n] = 0;
	DBG("DEBUG: cutter: billing_hangedup(): received '%s'.\n", buf.s);


	int code = bparser_code(buf.s);
	if (code < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"Invalid format of answer code.\n");
		return -1;
	}
	*result = code;

	char* msg = buf.s + 3; // 200 Descr=...


	str name, val;
	int pos = 0;
	int index;
	static char const* params[2] = {"Descr", NULL};

	r = bparser_parse(msg, params, &name, &val, &pos, &index);
	while (r == 1) {
		
		switch (index) {
		case -1:
			LOG(L_WARN, "WARNING: cutter: billing_hangedup(): "
			"unknown parameter in reply.\n");
			break;
		}
		r = bparser_parse(msg, params, &name, &val, &pos, &index);
	}
	if (r < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"invalid syntax of reply.\n");
		return -1;
	}

	if (billing_command("200 Descr=\"OK\"\n") < 0) {
		LOG(L_ERR, "ERROR: cutter: billing_hangedup(): "
		"write ack command failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: billing_hangedup(): billing result = %d.\n", *result);

	return 0;
}

