/*
 * Copyright (C) 2008 Petr Hruska <phru@ucw.cz> 
 * 
 * This file is part of Cutter module.
 * 
 * Cutter module is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <ctype.h>
#include <string.h>

#include "../../sr_module.h"

#include "bparser.h"

#define 	NOT_FOUND		-2
#define 	SYNTAX_ERROR		-1
#define 	NOTHING_TO_PARSE	0
#define 	PARSED_OK		1



inline static void skip_whitespaces(char const* msg, int* pos)
{
	while (msg[*pos] && (msg[*pos] == ' ' || msg[*pos] == '\t')) (*pos)++;
	
}

inline static int check_end(char const* msg, int* pos)
{
	if (!msg[*pos]) return PARSED_OK;

	return NOT_FOUND;

}


/* returns number of skipped characters [0-9A-Za-z] */
inline static int skip_name(char const* msg, int* pos)
{
	int r = 0;
	while (msg[*pos]) {

		int c = msg[*pos];
		if ((c < '0' || c > '9') && (c < 'A' || c > 'Z') && (c < 'a' || c > 'z')) break;
		r++;
		(*pos)++;
	}

	return r;
}

inline static int skip_char(int c, char const* msg, int* pos)
{
	if (msg[*pos] == c) {
		(*pos)++;
		return PARSED_OK;
	} else 
		return NOT_FOUND;
	
}


inline static int parse_string(char const* msg, int* pos, int* val_start, int* val_end)
{
	int escape = 0;
	int start = *pos;
	int len = 0;

	while (msg[*pos] && (escape || msg[*pos] != '"')) {

		if (escape) {
			int c = msg[*pos];
			if (c != 'n' && c != '"' && c != '\\') return SYNTAX_ERROR;
			escape = 0;
		} else {
			
			if (msg[*pos] == '\\') escape = 1;
		}
		(*pos)++;
		len++;
	}

	if (!msg[*pos]) return SYNTAX_ERROR;

	*val_start = start;
	*val_end = start + len;

	return PARSED_OK;
}


inline static int parse_unsigned(char const* msg, int* pos, int* val_start, int* val_end)
{
	int start = *pos;
	int len = 0;

	//DBG("DEBUG: cutter: parse_unsigned(): starting, pos=%d'.\n", *pos);

	while (msg[*pos]) {

		int c = msg[*pos];
		if (c < '0' || c > '9') break;
		len++;
		(*pos)++;
	}

	if (!len) return SYNTAX_ERROR;

	*val_start = start;
	*val_end = start + len;

	return PARSED_OK;
}


inline static int parse_signed(char const* msg, int* pos, int* val_start, int* val_end)
{
	int start = *pos;

	if (msg[*pos] == '-') (*pos)++;

	int r = parse_unsigned(msg, pos, val_start, val_end);
	if (r != PARSED_OK) return r;

	*val_start = start;
	return PARSED_OK;
		
}

int bparser_method_length(char const* msg)
{
	int pos = 0;
	skip_name(msg, &pos);
	//DBG("DEBUG: cutter: bparser_method_length(): pos = %d, msg='%s'.\n", pos, msg);
	return pos;
}

int bparser_code(char const* msg)
{
	int i;
	int len = bparser_method_length(msg);
	if (len != 3) return -1;

	int r = 0;
	for (i = 0; i < len; i++) {
		int c = msg[i];
		if (c < '0' || c > '9') return -1;
		c -= '0';
		r = 10 * r + c;
	}

	return r;
}

int bparser_parse(char* msg, char const* params[], str* name, str* val, int* pos, int* index)
{
	int name_start, name_end;
	int val_start, val_end;
	int i;
	int r;

	//DBG("DEBUG: cutter: bparser_parse(): starting, pos=%d, msg='%s'.\n", *pos, msg);


	*index = -1;

	skip_whitespaces(msg, pos);

	if (check_end(msg, pos) == PARSED_OK) {
		// end of msg
		return NOTHING_TO_PARSE;
	}

	name_start = *pos;
	r = skip_name(msg, pos);
	if (!r) return SYNTAX_ERROR; //empty name
	name_end = name_start + r;

	i = 0;
	while (params[i] && strncasecmp(params[i], msg + name_start, r)) i++;

	if (!params[i]) i = -1;

	skip_whitespaces(msg, pos);
	if (skip_char('=', msg, pos) != PARSED_OK) return SYNTAX_ERROR;
	skip_whitespaces(msg, pos);

	r = skip_char('"', msg, pos);
	if (r == PARSED_OK) {
		r = parse_string(msg, pos, &val_start, &val_end);
		if (r != PARSED_OK) return SYNTAX_ERROR;
		if (skip_char('"', msg, pos) != PARSED_OK) return SYNTAX_ERROR;
	} else {
		r = parse_signed(msg, pos, &val_start, &val_end);
		if (r != PARSED_OK) return SYNTAX_ERROR;
	}

	// if there is comma, skip it
	skip_whitespaces(msg, pos);
	if (skip_char(',', msg, pos) != PARSED_OK) {

		if (check_end(msg, pos) != PARSED_OK) {
			DBG("DEBUG: cutter: bparser_parse(): missing comma after parameter.\n");
			return SYNTAX_ERROR;
		}
	}

	name->s = msg + name_start;
	name->len = name_end - name_start;
	val->s = msg + val_start;
	val->len = val_end - val_start;
	*index = i;

	return PARSED_OK;
}


void bparser_unescape(str* s)
{
	int i;
	int j = 0;
	int escape = 0;
	for (i = 0; i < s->len; i++) {
	
		int c = s->s[i];

		if (escape) {
			switch (c) {
			case 'n': 
				s->s[j++] = '\n';
				break;
			default:
				s->s[j++] = c;

			}
			escape = 0;
		} else {
			if (c == '\\') 
				escape = 1;
			else 
				s->s[j++] = c;
		}
	}
	s->len = j;
}

