/*
 * Copyright (C) 2008 Petr Hruska <phru@ucw.cz> 
 * 
 * This file is part of Cutter module.
 * 
 * Cutter module is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "../../sr_module.h"
#include "../../mem/mem.h"
#include "../../parser/msg_parser.h"
#include "../../parser/parse_from.h"
#include "../../parser/parse_to.h"
#include "../../parser/contact/parse_contact.h"
#include "../../ut.h"
#include "../tm/tm_load.h"
#include "../../dset.h"


#include "billing.h"
#include "db_stuff.h"
#include "timer.h"
#include "fifo.h"
#include "misc.h"
#include "cut.h"
#include "cutter.h"
#include "rtpproxy.h"
 
MODULE_VERSION/*Module */

#define RET_OK		+1
#define RET_ERROR	-1	//something went wrong (db, billing, ...)
#define RET_FORBIDDEN	-2	//payment required

/*
 * Module parameters 
 * */


/* Database url */
char *db_url = 0;

/*
 * Functions
 * */
 
int mod_init(void);
void mod_destroy(void);
int child_init(int);

int cutter_universal(struct sip_msg* msg, char* uri, char* str2);

int billing_advance = 0;
void* bbuf;
 
/*
 *  * Script commands we export.
 *   */
cmd_export_t cmds[]={
	{"cutter",  cutter_universal,  0, NULL, REQUEST_ROUTE|ONREPLY_ROUTE },
	{0,0,0,0,0}
};
 
/*
 * Script parameters
 * */
param_export_t mod_params[] = {
	{"db_url",		STR_PARAM, &db_url		},
	{"db_table_name",	STR_PARAM, &cutter_db_table_name},
	{"billing_address",	STR_PARAM, &billing_address	},
	{"billing_timeout",	INT_PARAM, &billing_timeout	},
	{"billing_duration",	INT_PARAM, &billing_duration	},
	{"billing_advance",	INT_PARAM, &billing_advance	},
	{ 0,0,0 }
};

 
/*
 * Export the statistics we have
 * */
stat_export_t mod_stats[] = {
	{0,0,0}
};
 
struct module_exports exports = {
	"cutter",	/* module's name */
	cmds,		/* exported functions */
	mod_params,	/* param exports */
	mod_stats,	/* exported statistics */
	mod_init,	/* module initialization function */
	0,		/* reply processing function */
	mod_destroy,	/* Destroy function */
	child_init	/* per-child init function */
};

struct tm_binds tmb;


/* binds to the corresponding database module
 * returns 0 on success, -1 on error */
int cutter_db_bind(char* db_url)
{
	if (bind_dbmod(db_url, &cutter_dbf)<0){
		LOG(L_ERR, "ERROR: cutter: cutter_db_bind: bind_db failed\n");
		return -1;
	}

	/* Check database capabilities */
	if (!DB_CAPABILITY(cutter_dbf, DB_CAP_INSERT)) {
		LOG(L_ERR, "ERROR: cutter: cutter_db_bind: Database module "
		"does not implement insert function\n");
		return -1;
	}

	return 0;
}

int cutter_tm_bind()
{
	/* load the TM API */
	if (load_tm_api(&tmb)!=0) {
		LOG(L_ERR, "ERROR: cutter: cutter_tm_bind(): can't load TM API\n");
		return -1;
	}

	return 0;
}


int cutter_db_init()
{
	cutter_db_handle = cutter_dbf.init(db_url);
	if (!cutter_db_handle){
		LOG(L_ERR, "ERROR: cutter: cutter_db_init(): "
			"Unable to connect to the database\n");
		return -1;
	}

	if (cutter_dbf.use_table(cutter_db_handle, cutter_db_table_name) < 0) {
		LOG(L_ERR, "ERROR: cutter: cutter_db_init(): Error in use_table.\n");
		return -1;
	}

	return 0;
}


/**
 *  * @return 0 to continue to load the OpenSER, -1 to stop the loading
 *   * and abort OpenSER.
 *    */
int mod_init(void) {

	DBG("DEBUG: cutter: mod_init(): start\n");

	if (cutter_timer_init() < 0) {
		LOG(L_ERR, "ERROR: cutter: mod_init(): timer init failed.\n");
		return -1;
	}

	if (cutter_tm_bind() < 0) {
		LOG(L_ERR, "ERROR: cutter: mod_init(): transaction module bind failed.\n");
		return -1;
	}

	if (db_url && db_url[0]) {
		if (cutter_db_bind(db_url)<0) {
			LOG(L_ERR, "ERROR: cutter: mod_init(): cutter_db_init " 
			"failed... did you load a database module?\n");
			return -1;
		}
	} else {
		LOG(L_ERR, "ERROR: cutter: mod_init(): failed... did you setup db_url?\n");
		return -1;
	}

	/* Check and parse billing_address parameter */
	if (parse_billing_address(billing_address) < 0) {
		LOG(L_ERR, "ERROR: cutter: mod_init(): "
		"Invalid billing_address parameter.\n");
		return -1;
	}

	if (cutter_fifo_init() < 0) {
		LOG(L_ERR, "ERROR: cutter: mod_init(): "
		"Error initializing fifo server.\n");
		return -1;
	}

	DBG("DEBUG: cutter: mod_init(): successfull finish\n");
	return 0;
}
 
/**
 *  * Called only once when OpenSER is shuting down to clean up module
 *   * resources.
 *    */
void mod_destroy() {

	//QUESTION: is this function called for each child? we should dealloc bbuf somewhere
	DBG("DEBUG: cutter: mod_destroy(): successfull finish\n");
	return;
}
 
/**
 * The rank will be o for the main process calling this function,
 * or 1 through n for each listener process. The rank can have a negative
 * value if it is a special process calling the child init function.
 * Other then the listeners, the rank will equal one of these values:
 * * PROC_MAIN      0  Main ser process
 * * PROC_TIMER    -1  Timer attendant process 
 * * PROC_FIFO     -2  FIFO attendant process
 * * PROC_TCP_MAIN -4  TCP main process
 * * PROC_UNIXSOCK -5  Unix domain socket server processes
 *
 * If this function returns a nonzero value the loading of OpenSER will
 * stop.
 * */
int child_init(int rank) {
	
	DBG("DEBUG: cutter: child_init(): start\n");
	if (cutter_db_init(db_url) < 0) {
		LOG(L_ERR, "ERROR: cutter: child_init(): db init failed\n");
		return -1;
	}

	bbuf = billing_alloc();
	if (!bbuf) {
		LOG(L_ERR, "ERROR: cutter: child_init(): allocation of command buffer failed.\n");
		return -1;
	}

	DBG("DEBUG: cutter: child_init(): successfull finish\n");
	return 0;
}

// parses to and from 
static int parse_to_header(struct hdr_field *to)
{
	struct to_body* b;

	/* maybe the header is already parsed! */
	if (to->parsed)
		return 0;
	/* bad luck! :-( - we have to parse it */

	/* first, get some memory */
	b = pkg_malloc(sizeof(struct to_body));
	if (b == 0) {
		LOG(L_ERR, "ERROR: cutter: parse_to_header: out of pkg_memory\n");
		return -1;
	}

	/* now parse it!! */
	memset(b, 0, sizeof(struct to_body));
	parse_to(to->body.s,to->body.s+to->body.len+1,b);
	if (b->error == PARSE_ERROR) {
		LOG(L_ERR, "ERROR: cutter: parse_to_header: bad to header\n");
		free_to(b);
		return -1;
	}

	to->parsed = b;

	return 0;
}


static int prepare_contact(struct sip_msg* msg)
{
	if (parse_headers(msg, HDR_CONTACT_F, 0) < 0 || !msg->contact) {
		LOG(L_WARN, "WARNING: cutter: prepare_contact(): "
		"Missing contact header field.\n");
		return -1;
	}

	if (parse_contact(msg->contact) < 0) {

		LOG(L_WARN, "WARNING: cutter: prepare_contact(): "
		"Cannot parse contact header.\n");
		return -1;
	}

	return 0;
}


static int parse_common_headers(struct sip_msg* msg)
{
	if (parse_headers(msg, HDR_FROM_F|HDR_TO_F|HDR_CALLID_F|HDR_CSEQ_F, 0) < 0
	|| !msg->from || !msg->to || !msg->callid || !msg->cseq) {
		LOG(L_WARN, "WARNING: cutter: parse_common_headers(): "
		"Missing some header fields.\n");
		return -1;
	}

	// parsing from header
	if (parse_to_header(msg->from) < 0) {

		LOG(L_WARN, "WARNING: cutter: parse_common_headers(): "
		"Cannot parse from header.\n");
		return -1;
	}

	// parsing to header
	if (parse_to_header(msg->to) < 0) {

		LOG(L_WARN, "WARNING: cutter: parse_common_headers(): "
		"Cannot parse to header.\n");
		return -1;
	}

	return 0;
}


static int cutter_parse_rr(struct sip_msg* msg)
{
	// HDR_OTHER_F allways parses entire message.
	// This is neccessary, because there can be multiple Record-Route
	// header fields.
	if (parse_headers(msg, HDR_OTHER_F, 0) < 0) {
		LOG(L_WARN, "WARNING: cutter: parse_rr(): "
		"Can't parse record-route headers.\n");
		return -1;
	}

	return 0;
}


static int update_rr(int id, struct sip_msg* msg)
{
	int len = 0;
	str s;

	if (cutter_parse_rr(msg) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_rr(): Can't parse record-route headers fields.\n");
		return -1;
	}

	struct hdr_field* rr = msg->record_route;

	while (rr) {
		
		len += rr->body.len + 1; //place for comma
		rr = rr->sibling;
	}

	len--; // we don't need last comma
	s.s = pkg_malloc(len);
	if (!s.s) {
		LOG(L_ERR, "ERROR: cutter: update_rr(): pkg_malloc failed.\n");
		return -1;
	}

	s.len = 0;
	rr = msg->record_route;

	while (rr) {
		memcpy(s.s + s.len, rr->body.s, rr->body.len);
		s.len += rr->body.len;
		if (s.len < len) s.s[s.len++] = ',';
		rr = rr->sibling;
	}

	if (cutter_db_update_rr(id, s) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_rr(): db update failed.\n");
		return -1;
	}

	pkg_free(s.s);
	return 0;
}


static int rr_count(struct sip_msg* msg)
{
	struct hdr_field* rr = msg->record_route;
	int n = 0;

	while (rr) {
		n++;
		rr = rr->sibling;
	}

	return n;
}


static int update_cseq(struct dlg_short_t* info, struct sip_msg* msg, int direction)
{
	str cseq_str = ((struct cseq_body*) msg->cseq->parsed)->number;
	int cseq;

	if (str2sint(&cseq_str, &cseq) < 0) {
		
		LOG(L_ERR, "ERROR: cutter: update_cseq(): "
		"Invalid CSeq value (not a number).\n");
		return -1;
	}

	if (cutter_db_update_cseq(info->id, cseq, direction) < 0) {

		LOG(L_ERR, "ERROR: cutter: update_cseq(): db update failed.\n");
		return -1;
	}

	return 0;
}


int update_bye(struct sip_msg* msg, struct dlg_short_t* info)
{
	int r = 0;

	int scheduled = (info->S == STATUS_PENDING) && info->check_time;
	DBG("DEBUG: cutter: update_bye(): scheduled = %d (S = %d, check_time = %ld).\n", scheduled, info->S, info->check_time);
	info->S = STATUS_TERMINATED;

	time_t time_start = 0;
	time_t time_finish = time(NULL);

	if (cutter_db_update_dlg(info->id, info->S, time_start, time_finish) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_bye(): unable to update dialog.\n");
		r = -1;
	}

	// send HangedUp message to billing
	int result;
	str descr;
	descr.s = "BYE received";
	descr.len = strlen(descr.s);
	if (info->monitor != -2) {
		if (billing_hangedup(bbuf, &msg->callid->body, time(NULL), &descr, &result) < 0)
		{
			LOG(L_ERR, "ERROR: cutter: update_bye(): error sending HangedUp message to a billing.\n");
			r = -1;
		} else {

			if (result / 100 != 2) {
				// billing error
				DBG("DEBUG: cutter: update_bye(): Error code %d reported by a billing, terminating call.\n", result);
			}
		}
	}

	if (scheduled && cutter_unschedule(info) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_bye(): "
		"could not remove scheduled call check.\n");
	}

	return r;
}


int update_early(struct dlg_short_t* info)
{
	info->S = STATUS_EARLY;
	if (cutter_db_update_dlg(info->id, info->S, 0, 0) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_early(): unable to update dialog.\n");
		return -1;
	}
	return 0;
}

int terminate_call(int id)
{
	struct dlg_full_t info;
	
	DBG("DEBUG: cutter: terminate_call(): dialog id = %d.\n", id);

	if (cutter_db_dialog_full(id, &info, NULL) < 0) {
		LOG(L_ERR, "ERROR: cutter: terminate_call(): Can't fetch data from database.\n");
		return -1;
	}

	cutter_cut(&info);

	return 0;
}

int update_pending(struct sip_msg* msg, struct dlg_short_t* info)
{

	int r = 0;

	info->S = STATUS_PENDING;

	if (update_rr(info->id, msg) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_pending(): unable to update rr.\n");
		r = -1;
	}

	time_t time_start = 0;
	time_t time_finish = time(NULL);

	if (cutter_db_update_dlg(info->id, info->S, time_start, time_finish) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_pending(): unable to update dialog.\n");
		r = -1;
	}

	str from_tag = ((struct to_body*)msg->from->parsed)->tag_value;
	str to_tag = ((struct to_body*)msg->to->parsed)->tag_value;
	if (rtpp_start_timer(msg->callid->body, from_tag, to_tag) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_pending(): error starting rtpproxy timer.\n");
		r = -1;
	}

	DBG("DEBUG: cutter: update_pending(): info->monitor=%d.\n", info->monitor);
	if (info->monitor == -2) return r;

	// send Answered message to billing
	int result;
	if (billing_answered(bbuf, &msg->callid->body, time(NULL), &info->check_time, &info->monitor, &result) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_pending(): error sending Answered message to billing.\n");
		r = -1;
		if (terminate_call(info->id) < 0) 
			LOG(L_ERR, "ERROR: cutter: update_pending(): error terminating call.\n");
	} else {

		if (result / 100 != 2) {
			// billing error
			DBG("DEBUG: cutter: update_pending(): Error code %d reported by billing, terminating call.\n", result);
			if (terminate_call(info->id) < 0) {
				LOG(L_ERR, "ERROR: cutter: update_pending(): error terminating call.\n");
				r = -1;
			}
		} else {
			if (info->check_time > 0 && info->monitor >= 0) {

				// schedule check time
				if (cutter_schedule(info, info->monitor > 0 ? -billing_advance : 0) < 0) {
					LOG(L_ERR, "ERROR: cutter: update_pending(): "
					"could not schedule call check.\n");
				}

			}

			if (cutter_db_update_check(info->id, info->check_time, info->monitor) < 0) {
				LOG(L_ERR, "ERROR: cutter: update_pending(): "
				"could not update checking info in database.\n");
			}
		}
	}
	

	return r;
}


static inline int reply_descr(struct sip_msg* msg, str* descr)
{
	if (msg->first_line.type != SIP_REPLY) {
		descr->s = NULL;
		descr->len = 0;
		LOG(L_ERR, "ERROR: cutter: reply_descr(): not a reply message.\n");
		return -1;
	}
	
	descr->s = msg->first_line.u.reply.status.s;
	descr->len = (msg->first_line.u.reply.reason.s - descr->s) + msg->first_line.u.reply.reason.len;

	return 0;
}


int update_failed(struct sip_msg* msg, struct dlg_short_t* info)
{
	int r = 0;

	info->S = STATUS_FAILED;

	if (cutter_db_update_dlg(info->id, info->S, 0, 0) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_failed(): unable to update dialog.\n");
		r = -1;
	}

	DBG("DEBUG: cutter: update_failed(): info->monitor=%d.\n", info->monitor);
	if (info->monitor == -2) return r;

	// send Cancelled message to billing
	int result;
	str descr;
	if (reply_descr(msg, &descr) < 0) {
		LOG(L_ERR, "ERROR: cutter: update_failed(): can't get message description.\n");
		r = -1;
	}

	if (billing_cancelled(bbuf, &msg->callid->body, time(NULL), &descr, &result) < 0)
	{
		LOG(L_ERR, "ERROR: cutter: update_failed(): error sending Cancelled message to billing.\n");
		r = -1;
	} else {

		if (result / 100 != 2) {
			// billing error
			DBG("DEBUG: cutter: update_failed(): Error code %d reported by billing.\n", result);
		}
	}
	

	return r;
}


int update_dlg_status(struct dlg_short_t* info, struct sip_msg* msg, int direction)
{
	// send back error reply
	if (info->S == STATUS_FORBIDDEN) return 0;

	if (msg->first_line.type == SIP_REQUEST) {

		int r = 0; //result
		if (update_cseq(info, msg, direction) < 0) {

			LOG(L_ERR, "ERROR: cutter: update_dlg_status(): cseq update failed.\n");
			r = -1;
			//do not return
		}

		switch (msg->first_line.u.request.method_value) {
		
		case METHOD_BYE:
			if (info->S == STATUS_TERMINATED) break;
			int rr = update_bye(msg, info);
			if (!r && rr < 0) r = rr;
			return r;
		}

	} else { //REPLY

		//check only replies to INVITE
		if (((struct cseq_body*) msg->cseq->parsed)->method_id == METHOD_INVITE)
		switch (msg->first_line.u.reply.statuscode / 100) {

		case 1:
			if (info->S != STATUS_START || info->S == STATUS_EARLY) break;
			return update_early(info);
		case 2:
			if (info->S == STATUS_PENDING) break;
			return update_pending(msg, info);
		case 3:
		case 4:
		case 5:
		case 6:
			DBG("DEBUG: cutter: update_dlg_status(): error reply to INVITE detected, status=%d.\n", info->S);
			if (info->S != STATUS_START && info->S != STATUS_EARLY) break;
			return update_failed(msg, info);
			break;
		}
	}

	// the message is not interresting
	return 0;
}

int snprint_ip(char* buf, size_t size, struct ip_addr* ip)
{
	switch(ip->af){
	case AF_INET:
		return snprintf(buf, size, "%d.%d.%d.%d", ip->u.addr[0],
			ip->u.addr[1], ip->u.addr[2], ip->u.addr[3]);
	case AF_INET6:
		return snprintf(buf, size, "%x:%x:%x:%x:%x:%x:%x:%x",
			htons(ip->u.addr16[0]), htons(ip->u.addr16[1]),
			htons(ip->u.addr16[2]), htons(ip->u.addr16[3]),
			htons(ip->u.addr16[4]), htons(ip->u.addr16[5]),
			htons(ip->u.addr16[6]), htons(ip->u.addr16[7]));
	default:
		LOG(L_ERR, "ERROR: cutter: snprint_ip(): Warning unknown address family %d.\n", ip->af);
		return -1;
        }
}



int handle_first_invite(struct sip_msg* msg, struct dlg_short_t* short_info)
{
	struct dlg_full_t info;
	str empty = { NULL, 0 };
	int result;
	str callcon;
	str srcip;
	const int max_addr_len = 24+1;
	char addrbuf[max_addr_len];

	DBG("DEBUG: cutter: handle_first_invite(): start\n");

	if (cutter_parse_rr(msg) < 0) {

		LOG(L_WARN, "WARNING: cutter: handle_first_invite(): "
		"Cannot parse record route headers.\n");
		return -1;
	}

	str cseq_str = ((struct cseq_body*) msg->cseq->parsed)->number;
	int cseq;

	if (str2sint(&cseq_str, &cseq) < 0) {
		
		LOG(L_ERR, "ERROR: cutter: handle_first_invite(): "
		"Invalid CSeq value (not a number).\n");
		return -1;
	}

	// parsing contact
	if (prepare_contact(msg) < 0) {

		LOG(L_WARN, "WARNING: cutter: handle_first_invite(): "
		"Cannot parse contact header.\n");
		return -1;
	}


	info.from_uri = ((struct to_body*)msg->from->parsed)->uri;
	info.to_uri = ((struct to_body*)msg->to->parsed)->uri;
	info.from_tag = ((struct to_body*)msg->from->parsed)->tag_value;
	info.to_tag = ((struct to_body*)msg->to->parsed)->tag_value;
	info.callid = msg->callid->body;
	info.rr = empty;
	info.n_hops = rr_count(msg);
	info.from_cseq = cseq;
	info.to_cseq = -1;
	info.from_hfb = ((struct to_body*) msg->from->parsed)->body;
	info.to_hfb = empty;
	info.time_start = 0;
	info.time_finish = 0;


	int r = snprint_ip(addrbuf, max_addr_len, &msg->rcv.src_ip);
	if (r >= max_addr_len || r < 0) {
		LOG(L_ERR, "ERROR: cutter: handle_first_invite(): Error composing srcip.\n");
		return -1;
	}
	srcip.s = addrbuf;
	srcip.len = strlen(addrbuf);


	info.from_contact = ((struct contact_body*)msg->contact->parsed)->contacts[0].uri;

	// this will be updated by a contact from reply
	get_request_uri(msg, &info.to_contact);

	// default values, which can be changed by billing server
	info.check_time = 0; //no querying
	info.monitor = 0;
	info.S = STATUS_START;



	if (billing_calling(bbuf, &info.callid, &info.from_uri, &info.to_contact, &srcip, &callcon, &info.check_time, &info.monitor, &result) < 0) {
		LOG(L_WARN, "WARNING: cutter: handle_first_invite(): Error while reporting new call to billing.\n");
		info.S = STATUS_FORBIDDEN;
	} else {


		if (result / 100 != 2) { // billing error
			
			info.S = STATUS_FORBIDDEN;
			DBG("DEBUG: cutter: handle_first_invite(): Non-2xx reply received from billing.\n");

		} else if (serstrcmp(&info.to_contact, &callcon)) {
			DBG("DEBUG: cutter: handle_first_invite(): rewriting ruri.\n");
			if (rewrite_uri(msg, &callcon) < 0) {
				info.S = STATUS_FORBIDDEN;
				LOG(L_ERR, "ERROR: cutter: handle_first_invite(): Can't rewrite callcon.\n");
			} else
				info.to_contact = callcon;
		}
	}

	// insert into database
	short_info->id = cutter_db_insert_dlg(&info);
	if (short_info->id < 0) {
		LOG(L_ERR, "ERROR: cutter: handle_first_invite(): "
		"Error while inserting a dialog to database\n");
		return -1;
	}

	short_info->check_time = info.check_time;
	short_info->monitor = info.monitor;
	short_info->S = info.S;

	return 0;
}


int handle_first_reply(struct sip_msg* msg, struct dlg_short_t* si)
{
	str callid, from_tag, to_tag;
	int direction;
	struct dlg_full_t info;

	// we need from_contact, let's find it
	callid = msg->callid->body;
	from_tag = ((struct to_body*)msg->from->parsed)->tag_value;
	to_tag.s = NULL;
	to_tag.len = 0;

	if (cutter_db_dialog_short(callid, from_tag, to_tag, si, &direction) < 0) {
		LOG(L_WARN, "WARNING: cutter: handle_first_reply(): Parent dialog not found.\n");
		return -1;
	}

	if (direction == -1) {
		LOG(L_WARN, "WARNING: cutter: handle_first_reply(): Wrong direction of first reply.\n");
		return -1;
	}

	if (cutter_db_dialog_full(si->id, &info, NULL) < 0) {
		LOG(L_WARN, "WARNING: cutter: handle_first_reply(): Fetch of full info failed.\n");
		return -1;
	}

	info.to_hfb = ((struct to_body*) msg->to->parsed)->body;
	info.to_tag = ((struct to_body*)msg->to->parsed)->tag_value;

	if (msg->first_line.u.reply.statuscode / 100 <= 2) {
		// parsing contact
		if (prepare_contact(msg) < 0) {

			LOG(L_WARN, "WARNING: cutter: handle_first_invite(): "
			"Cannot parse contact header.\n");
			cutter_db_free_result(&info);
			return -1;
		}
		info.to_contact = ((struct contact_body*)msg->contact->parsed)->contacts[0].uri;
	}

	//all necessary information is collected
	si->id = cutter_db_insert_dlg(&info);
	if (si->id < 0) {

		LOG(L_ERR, "ERROR: cutter: handle_first_reply(): "
		"Error while inserting a dialog to database\n");
		cutter_db_free_result(&info);
		return -1;
	}

	si->S = info.S;
	/* This should be STATUS_START or STATUS_EARLY.
	 * STATUS_START is initially set when creating the parent record of
	 * INVITE request. This can be updated to STATUS_EARLY by Trying
	 * response, which MAY have empty To tag just like INVITE.
	 */
	

	si->check_time = info.check_time;
	si->monitor = info.monitor;
	cutter_db_free_result(&info);


	/* This function will update si->S and setup the database. Possible new
	 * status is EARLY (180 Ringing), or PENDING (200 OK).
	 */
	if (update_dlg_status(si, msg, -1) < 0) {

		LOG(L_ERR, "ERROR: cutter: handle_first_reply(): Error occured while updating dialog status.\n");
		return -1;

	}


	return 0;
}


/*
 * info is uninitialized structure, the dialog record is not present in
 * a database.
 *
 * -1 => error
 * 0 => ok, done
 * +1 => non-dialog message, nothing to do
 */
int create_new_dialog(struct sip_msg *msg, struct dlg_short_t* info)
{

	// let's check for the first invite
	if (msg->first_line.type == SIP_REQUEST) {

		DBG("DEBUG: cutter: create_new_dialog(): request %.*s\n",
			msg->first_line.u.request.method.len, msg->first_line.u.request.method.s);

		if (msg->first_line.u.request.method_value == METHOD_INVITE) {
			if (handle_first_invite(msg, info) < 0) {
				LOG(L_ERR, "ERROR: cutter: create_new_dialog(): "
				"Error while processing first invite.\n");
				return -1;
			}

		} else {
			// non-dialog requests, like REGISTER
			DBG("DEBUG: cutter: create_new_dialog(): non-dialog request.\n");
			return 1;
		}
	} else if (msg->first_line.type == SIP_REPLY) {


		/* First reply was not found in a database, because called side sets up
		 * To tag, which identifies dialog. 
		 */
		DBG("DEBUG: cutter: create_new_dialog(): reply %d %.*s\n",
			msg->first_line.u.reply.statuscode, 
			msg->first_line.u.reply.reason.len, msg->first_line.u.reply.reason.s);

			if (handle_first_reply(msg, info) < 0) {
			LOG(L_WARN, "WARNING: cutter: create_new_dialog(): "
			"Error while processing reply within unknown dialog.\n");
			return -1;
		}

	} else {
		DBG("DEBUG: cutter: create_new_dialog(): invalid message type.\n");
		return -1;
	}

	return 0;
}


int cutter_universal(struct sip_msg* msg, char* uri, char* str2)
{
	struct dlg_short_t info;
	str from_tag, to_tag;
	int direction = 0; // +1 is direction of initial invite

	int n; //control variable
	// n < 0 => error
	// n == 0 // dialog found or created
	// n > 0 not a dialog message

	DBG("=== DEBUG: CUTTER: cutter_universal() start\n");

	// parse callid, from tag and to tag
	if (parse_common_headers(msg) < 0) {
		LOG(L_ERR, "ERROR: cutter: cutter_universal(): "
		"Cannot parse common headers.\n");
		return RET_ERROR;
	}

	from_tag = ((struct to_body*)msg->from->parsed)->tag_value;
	to_tag = ((struct to_body*)msg->to->parsed)->tag_value;

	// lookup for an existing dialog record in the database
	int r = cutter_db_dialog_short(msg->callid->body, from_tag, to_tag, &info, &direction);
	if (r == -1) {
		LOG(L_ERR, "ERROR: cutter: cutter_universal(): Error accessing database.\n");
		return RET_ERROR;
	}

	// not found
	if (r == -2) {
		n = create_new_dialog(msg, &info);
		if (n < 0) {
			LOG(L_ERR, "ERROR: cutter: cutter_universal(): Error creating dialog.\n");
			return RET_ERROR;
		} else if (n == 0) {
			switch (info.S) {
			case STATUS_FORBIDDEN:
				return RET_FORBIDDEN;
			case STATUS_START:
			case STATUS_EARLY:
			case STATUS_PENDING:
				return RET_OK;
			default:
				LOG(L_CRIT, "CRITICAL: cutter: cutter_universal(): Internal error.\n");
				return RET_ERROR;
			}
		} else {
			DBG("DEBUG: cutter: cutter_universal(): finish, message not in dialog.\n");
			return RET_OK;
		}
	}

	DBG("DEBUG: cutter: cutter_universal(): found dialog.\n");
	// replies has same from and to tags as their requests
	if (msg->first_line.type == SIP_REPLY) direction *= -1;

	// write messages to db, etc.
	if (update_dlg_status(&info, msg, direction) < 0) {

		LOG(L_ERR, "ERROR: cutter: cutter_universal(): Error occured while updating dialog status.\n");
		return RET_ERROR;
	}

	if (info.S == STATUS_FORBIDDEN) {
		DBG("DEBUG: cutter: cutter_universal(): finish, forbidden.\n");
		return RET_FORBIDDEN;
	}

	DBG("DEBUG: cutter: cutter_universal(): finish, ok.\n");
	return RET_OK;
}
