/*
 * Copyright (C) 2008 Petr Hruska <phru@ucw.cz> 
 * 
 * This file is part of Cutter module.
 * 
 * Cutter module is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "../../sr_module.h"
#include "../../mem/shm_mem.h"
#include "../../lock_ops.h"
#include "timer_data.h"

// current size of struct timer_data_t is 16B on i386 architecture,
// this gives us 5000*16=80KB size of largest table
#define MAX_RECORDS	5000


struct timer_data_t {
	
	struct timer_item_t ti;

	// circular linked list of items belonging to one hash value, 
	// or free (not allocated) items
	struct timer_data_t* next;
	struct timer_data_t* prev;

	int hindex; // index to hash table
};

static struct timer_data_t *data;
static struct timer_data_t* *heap;
static struct timer_data_t* *htable;
static struct timer_data_t* *available; //dealocated items
static int *td_cnt; // number of allocated records

gen_lock_t * td_lock; // synchronization of access


//
// **** misc functions
//


void td_link(struct timer_data_t** list, struct timer_data_t* item)
{
	//empty list
	if (!*list) {
		*list = item;
		item->next = item;
		item->prev = item;
		return;
	}

	item->next = *list;
	item->prev = (*list)->prev;
	(*list)->prev = item;
	item->prev->next = item;
}


void td_unlink(struct timer_data_t** list, struct timer_data_t* item)
{
	if (*list == item) {

		if (item->next == item) {
			// last item
			*list = NULL;
			return;
		} else 
			*list = item->next;
	}

	// *list != item

	item->next->prev = item->prev;
	item->prev->next = item->next;
}

// allocator

struct timer_data_t* td_alloc()
{
	struct timer_data_t* r;

	if (!*available) return NULL;

	if ((*available)->next == *available) {
		// last item
		r = *available;
		*available = NULL;
		return r;
	}

	r = *available;
	*available = r->next;
	r->next->prev = r->prev;
	r->prev->next = r->next;

	return r;
}


void td_free(struct timer_data_t *td)
{
	td_link(available, td);
}


//
// hash table 
//

// returns index 0 <= t < MAX_RECORDS based on dialog id
static int htable_hash(int dlg)
{
	unsigned int t;
	t = (unsigned int) dlg;
	t = (t * 5713 + 158) % MAX_RECORDS;
	return (int) t;
}

static void htable_insert(struct timer_data_t* td)
{
	int h;
	h = htable_hash(td->ti.dlg);
	td_link(&htable[h], td);
}


struct timer_data_t* htable_remove(int dlg)
{
	struct timer_data_t* td;
	int h;

	h = htable_hash(dlg);
	td = htable[h];
	if (!td) return NULL;

	// check first element
	if (td->ti.dlg == dlg) {
		td_unlink(&htable[h], td);
		return td;
	}
	
	// move on and check rest
	td = td->next;
	while (td != htable[h] && td->ti.dlg == dlg) td = td->next;

	// we stoped on first element, there was no match
	if (td == htable[h]) return NULL;

	// found
	td_unlink(&htable[h], td);
	return td;
}




//
// heap
//
//


inline static int heap_cmp(int a, int b)
{
	if (heap[a]->ti.t < heap[b]->ti.t) return -1;
	if (heap[a]->ti.t > heap[b]->ti.t) return +1;
	return 0;
}


inline static void heap_swap(int a, int b)
{
	struct timer_data_t* tmp;

	tmp = heap[a];
	heap[a] = heap[b];
	heap[b] = tmp;
}


static void heap_up(int a)
{
	int p; // parent
	p = (a - 1) / 2;

	while (a > 0 && heap_cmp(p, a) > 0) {

		heap_swap(p, a); // fix heap condition
		heap[a]->hindex = a;

		a = p; // up one level
		p = (a - 1) / 2;
	}

	heap[a]->hindex = a;
}


static void heap_down(int a)
{
	int s1, s2; // sons
	int s; // son with minimum time (s = argmin(s1, s2))
	s1 = 2*a + 1;
	s2 = s1 + 1;
	// while both sons are valid
	while (s2 < *td_cnt) {

		if (heap_cmp(a, s1) > 0) {
			//a > s1
			if (heap_cmp(s1, s2) > 0) 
				// s1 > s2
				s = s2;
			else
				// s1 < s2
				s = s1;
		} else if (heap_cmp(a, s2) > 0) 
			// a <= s1 && a > s2 ----> s1 > s2
			s = s2;
		else
			break;
		
		heap_swap(a, s);
		heap[a]->hindex = a;
		a = s;

		s1 = 2*a + 1;
		s2 = s1 + 1;
	}

	if (s1 == *td_cnt - 1) {
		// s1 is only valid son
		if (heap_cmp(a, s1) > 0) {
			
			heap_swap(a, s1);
			heap[a]->hindex = a;
			a = s1;
		}
	}

	heap[a]->hindex = a;
}


static int heap_insert(struct timer_data_t* td)
{
	if (*td_cnt == MAX_RECORDS) return -1;

	heap[(*td_cnt)++] = td;
	heap_up(*td_cnt - 1);

	return 0;
}

/*
static void heap_check(char* s)
{
	int i;
	for (i = 0; i < *td_cnt; i++) if (heap[i]->hindex != i) DBG("ERROR i = %d, hindex = %d\n", i, heap[i]->hindex);
	for (i = 0; i < *td_cnt; i++) 
	{
		if (heap_cmp((i - 1) / 2, i) > 0) DBG("ERROR %s, parent of %d is greater.\n", s, i);
	}
}
*/

static int heap_remove(struct timer_data_t* td)
{
	int a;

	a = td->hindex;

	// we will replace element on place a with 
	// element on place *td_cnt - 1. Let's check
	// how the heap condition can be broken.
	if (heap_cmp(a, *td_cnt - 1) > 0) {
		// new element is smaller  (val[a] > val[*td_cnt - 1])
		// subtree is ok, but parent can be greater now
		heap[a] = heap[--*td_cnt];
		heap_up(a);
	} else {
		heap[a] = heap[--*td_cnt];
		heap_down(a);
	}

	return 0;
}

static inline struct timer_data_t* heap_top()
{
	if (*td_cnt) 
		return heap[0]; 
	else
		return NULL;
}

static int td_init_alloc()
{
	// allocation of data structures in shared memory
	data = shm_malloc(sizeof(*data) * MAX_RECORDS);
	if (!data) {
		LOG(L_ERR, "ERROR: cutter: td_init_alloc(): shm_malloc failed (data)\n");
		return -1;
	}
	heap = shm_malloc(sizeof(*heap) * MAX_RECORDS);
	if (!heap) {
		LOG(L_ERR, "ERROR: cutter: td_init_alloc(): shm_malloc failed (heap)\n");
		return -1;
	}
	htable = shm_malloc(sizeof(*htable) * MAX_RECORDS);
	if (!htable) {
		LOG(L_ERR, "ERROR: cutter: td_init_alloc(): shm_malloc failed (htable)\n");
		return -1;
	}
	available = shm_malloc(sizeof(*available));
	if (!available) {
		LOG(L_ERR, "ERROR: cutter: td_init_alloc(): shm_malloc failed (available)\n");
		return -1;
	}
	td_cnt = shm_malloc(sizeof(*td_cnt));
	if (!td_cnt) {
		LOG(L_ERR, "ERROR: cutter: td_init_alloc(): shm_malloc failed (td_cnt)\n");
		return -1;
	}

	return 0;
}

static int td_lock_init()
{
        td_lock = shm_malloc(sizeof(gen_lock_t));

        if (!td_lock) {
                LOG(L_CRIT, "ERROR: cutter: td_lock_init(): could not allocate lock.\n");
                return -1;
        }
        if (!lock_init(mem_lock)) {

                LOG(L_CRIT, "ERROR: cutter: td_lock_init(): could not initialize lock.\n");
                return -1;
        }

	return 0;
}

//
// **** interface functions
//

int td_init()
{
	int i;

	if (td_init_alloc() < 0) {
		LOG(L_ERR, "ERROR: cutter: td_init(): memory allocation failure.\n");
		return -1;
	}

	if (td_lock_init() < 0) {
		LOG(L_ERR, "ERROR: cutter: td_init(): lock allocation failure.\n");
		return -1;
	}

	*td_cnt = 0; // this initializes heap

	// create linked list of all items
	// fill hash table
	for (i = 0; i < MAX_RECORDS; i++) {
		
		data[i].next = &data[(i + 1) % MAX_RECORDS];
		data[i].prev = &data[(i - 1 + MAX_RECORDS) % MAX_RECORDS];

		htable[i] = NULL;
	}

	*available = &data[0];

	return 0;
}


static int td_insert_unsafe(struct timer_item_t* pti)
{
	struct timer_data_t* td;

	td = td_alloc();
	if (!td) {
		LOG(L_ERR, "ERROR: cutter(): td_insert(): Maximum concurrent calls exceeded.\n");
		return -1;
	}

	td->ti = *pti;
	
	htable_insert(td);

	if (heap_insert(td) < 0) {
		//maximum number of records should be captured by alloc, this should never happen
		LOG(L_ERR, "ERROR: cutter(): td_insert(): Invalid heap operation.\n");
		return -1;
	};

	return 0;
}


int td_insert(struct timer_item_t* pti)
{
	int r;

	lock_get(td_lock);
	r = td_insert_unsafe(pti);
	lock_release(td_lock);

	return r;
}


static int td_remove_unsafe(int dlg)
{
	struct timer_data_t* td;
	
	td = htable_remove(dlg);
	if (!td) return -2;

	if (heap_remove(td) < 0) {
		LOG(L_ERR, "ERROR: cutter(): td_remove(): Invalid heap operation.\n");
		return -1;
	}

	td_free(td);

	return 0;
}


int td_remove(int dlg)
{
	int r;

	lock_get(td_lock);
	r = td_remove_unsafe(dlg);
	lock_release(td_lock);

	return r;
}


//removes the item on top and fills *pti
static int td_extract_unsafe(struct timer_item_t* pti)
{
	struct timer_data_t* td;

	td = heap_top();
	if (!td) return -2;

	*pti = td->ti;
	return td_remove_unsafe(pti->dlg);
}


int td_extract(struct timer_item_t* pti)
{
	int r;

	lock_get(td_lock);
	r = td_extract_unsafe(pti);
	lock_release(td_lock);

	return r;
}


int td_extract_elapsed(struct timer_item_t* pti, time_t t)
{
	int r;
	struct timer_data_t* td;

	r = -2;
	lock_get(td_lock);


	td = heap_top();
	if (td && td->ti.t <= t) r = td_extract_unsafe(pti);

	lock_release(td_lock);

	return r;
}
